"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const _ = require("lodash");
const helpers_1 = require("../helpers");
/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Ventures Pty Ltd. All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */
class CRUDMiddleware {
    constructor(store_plugin) {
        this.name = "generic_data_store";
        this.method_actions = {
            "GET": "read", "POST": "create", "PUT": "update", "DELETE": "delete", "LIST": "list"
        };
        this.store = store_plugin;
        if (store_plugin.name.indexOf(".") > 0) {
            this.name = store_plugin.name;
        }
        else {
            this.name = "api." + store_plugin.name;
        }
        store_plugin.context.log({ code: "api:crud:middleware", name: this.name, plugin: store_plugin.name });
    }
    /**
     * fn() is the generic entry point to initialize a plugin.
     *
     * @param context
     * @param options
     */
    fn(operation, _options) {
        let store = this.store;
        let self = this;
        // merge plugin options, middleware options and request parameters ..
        let options = _.extend({}, store.options, _options);
        return function (req, res, next) {
            assert(req && res && next, "invalid middleware");
            // get actionId from 'features' then method lookup, defaults to the operation's actionId.
            let actionId = options.actionId || self.method_actions[operation.actionId.toUpperCase()] || operation.actionId;
            // heuristic map infer 'list' versus 'read'
            if (actionId == "read" && (operation.resource.endsWith("/") || req.params.length == 0)) {
                actionId = "list";
            }
            // create a model from the request's JSON payload
            let model = _.extend({}, req.body);
            // closure to handle success / failure of the CRUD operation  
            let response_handler = function (err, r) {
                if (err) {
                    operation.context.warn(err.code ? err : { code: "api:crud:failed:" + actionId, message: err,
                        operationId: operation.operationId,
                        actionId: actionId,
                        options: options,
                        model: model._id });
                    res.status(err.statusCode || 400); // 400 bad request
                    res.send(err.code ? err : { code: err.code, message: err.err, model: model._id });
                }
                else {
                    // broadcast local event
                    operation.context.bus.emit("api:crud:" + actionId, { model: model, options: options });
                    // remove circular references
                    r = helpers_1.Vars.decycle(r);
                    // send response
                    res.status(200);
                    res.send(r);
                }
            };
            try {
                self.handle(operation, actionId, model, req, options, response_handler);
            }
            catch (err) {
                console.error(err);
                let _error = {
                    code: "api:crud:error:" + actionId,
                    error: err,
                    statusCode: 500,
                    operationId: operation.operationId,
                    actionId: actionId,
                    options: options,
                    model: model._id
                };
                response_handler(_error, null);
            }
        };
    }
    handle(operation, actionId, model, req, _options, response_handler) {
        let store = this.store;
        // get request & query params based on OASv3 spec
        let req_param_keys = operation.parameterKeys("path");
        let req_query_keys = operation.parameterKeys("query");
        // merge params - path params take precedence over query params.
        let merged_params = _.extend({}, req.query, req.params);
        // filtering - pick based on OASv3 parameters
        let filter = _.extend({}, _.pick(req.query, req_query_keys), _.pick(req.params, req_param_keys));
        let filterBy = operation.feature.filterBy || false;
        filter = filterBy ? _.pick(filter, filterBy) : filter;
        // TODO: query model based on body paramsters
        // let model_id = model[store.getIDField()] || filter[store.getIDField()];
        let options = _.extend({}, _.pick(this.store.options, ["database", "collection"]), _options);
        store.context.log({ code: "api:crud:handle", name: this.name, operationId: operation.operationId, actionId: actionId, store: store.name, options: options, filter: filter, params: merged_params });
        // console.log("CRUD [%s]: %o -> %o [keys: q: %o --> p: %o]", actionId, merged_params, options, req_query_keys, req_param_keys);
        // perform the CRUD facade operation
        switch (actionId) {
            case "create":
                store.create(operation, model, merged_params, options, response_handler);
                break;
            case "read":
                store.read(operation, filter, merged_params, options, response_handler);
                break;
            case "list":
                store.list(operation, filter, merged_params, options, response_handler);
                break;
            case "update":
                store.update(operation, model, merged_params, options, response_handler);
                break;
            case "delete":
                store.delete(operation, model, merged_params, options, response_handler);
                break;
            case "health":
            case "healthz":
                store.healthy(options, response_handler);
                break;
            default:
                // polymorphic heuristics
                let crud_fn = store[actionId];
                if (crud_fn) {
                    crud_fn(model, merged_params, response_handler);
                }
                else {
                    let _error = { code: "crud:operationId:unknown", operationId: operation.operationId, actionId: actionId };
                    response_handler(_error);
                }
                break;
        }
    }
}
exports.CRUDMiddleware = CRUDMiddleware;
//# sourceMappingURL=CRUDMiddleware.js.map