"use strict";
/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Ventures Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const assert = require("assert");
const express = require("express");
const Logger = require("./Logger");
const uuidv5 = require("uuid/v5");
require('source-map-support').install();
const EventEmitter = require("events");
const helpers_1 = require("../helpers/");
const Registry_1 = require("./Registry");
require("source-map-support/register");
const path = require("path");
const Debug = require("debug");
const NODE_ENV = process.env.NODE_ENV || "development";
const IS_DEV_ENV = (NODE_ENV == "development");
class Chassis {
    constructor(config, bootFn) {
        this.plugins = new Registry_1.Registry();
        this.middleware = new Registry_1.Registry();
        this.booted = false;
        this.config = config;
        this.context = this.init(config);
        this.boot(bootFn);
    }
    ;
    /**
     * Launch the Chassis.

     * Initializes the Chassis (process event listeners, etc)
     *
     * Installs all of the registered plugins.
     *
     * Starts listening for HTTP/S requests on a port.
     *
     * Emits "api:init" event to announce it's availability
     *
     * @returns {IChassisContext}
     */
    start() {
        let context = this.context; // unstructured - so we can mutute our context
        assert(!context.server, "already started");
        assert(context.api, "missing context.api");
        assert(context.api.get, "missing context.api.get");
        assert(context.api.listen, "missing context.api.listen");
        assert(context.logger, "missing context.logger");
        assert(context.config, "missing context.config");
        assert(context.config.name, "missing context.config.name");
        let config = context.config;
        let port = context.api.get('port');
        assert(port, "context.api port not found");
        this.installPlugins(this.context, this.config);
        let _started = {
            code: "api:start",
            name: this.config.name,
            port: this.config.port,
            package: context.pkg.name || false,
            version: context.pkg.version || false,
            features: _.keys(this.config.features),
            plugins: _.keys(this.context.plugins.features),
            middleware: _.keys(this.context.middleware.features),
            environment: NODE_ENV
        };
        // process exceptions
        config.process_guard && this.process_guard();
        // start lisenting for HTTP(S) requests on port
        context.server = context.api.listen(port, function () {
            // announce Chassis state
            context.log(_started);
            context.bus.emit(_started.code, { _started });
            // for humans
            console.log("%s v%s on port %s (%s)", _started.package, _started.version, config.port, NODE_ENV);
        });
        return context;
    }
    ;
    pkg() {
        let pkg = null;
        try {
            pkg = helpers_1.Vars.load(process.cwd() + "/package.json");
        }
        catch (e) {
            console.log("suggest: add package.json");
            let name = path.dirname(process.cwd()).split(path.sep).pop();
            pkg = { name: name, version: "0.0.0" };
        }
        return pkg;
    }
    /**
     * Initialize a ChassisContext from a given configuration object
     *
     * Create the IChassisContext and binds the runtime instances
     *
     * Must be called before start()
     *
     * @param config
     * @returns {IChassisContext}
     */
    init(config) {
        assert(config.name, "missing config.name");
        assert(config.port, "missing config.port");
        let debug = Debug(config.name);
        helpers_1.Vars.resolve_externals(config); // resolve @ references
        // generate a unique UUID
        let uuid = uuidv5(config.name + "_" + config.port + "_" + Date.now(), uuidv5.DNS);
        // define logging sub-system for re-use across project
        config.logging = _.extend({}, config.logging);
        let logger = Logger.define_logger(config.logging);
        assert(logger, "invalid logger");
        // define audit logs
        config.auditing = _.extend({}, config.auditing);
        let auditor = Logger.define_logger(config.auditing);
        assert(auditor, "invalid auditor");
        // we're event driven by default
        let bus = new EventEmitter();
        // initialize express - to serve our HTTP/S requests
        let apis = express();
        // define the port that Express will listen on
        apis.set('port', config.port);
        //TODO: security harden
        apis.disable('x-powered-by');
        // https://expressjs.com/en/advanced/best-practice-security.html
        bus.on("error", function (err) {
            console.error(err);
            logger.error({ code: "api:bus:error", message: "event error: " + err, stack: err.stack || null });
        });
        // create our Chassis context - bind our instances into the object
        let pkg = this.pkg();
        let self = this;
        let context = {
            uuid: uuid,
            bus: bus,
            api: apis,
            pkg: pkg,
            resolve: (item) => {
                return helpers_1.Vars.resolve_externals(item);
            },
            config: config,
            // openapi: null,
            secrets: null,
            logger: logger,
            auditor: auditor,
            plugins: this.plugins,
            middleware: this.middleware,
            metrics: null,
            error: function (_log) {
                assert(_log, "missing error.message");
                _log.message = _log.message || _log.code;
                self._stamp(context, _log);
                debug(_log.code);
                bus.emit("api:error", _log);
                logger.error(_log);
                return true;
            },
            warn: function (_log) {
                assert(_log, "missing log.message");
                _log.message = _log.message || _log.code;
                self._stamp(context, _log);
                debug(_log.code);
                bus.emit("api:warn", _log);
                logger.warn(_log);
                return true;
            },
            log: function (_log) {
                assert(_log, "missing log.message");
                _log.message = _log.message || _log.code;
                self._stamp(context, _log);
                debug(_log.code);
                logger.info(_log);
                return true;
            },
            trace: function (_log) {
                assert(_log, "missing log.message");
                _log.message = _log.message || _log.code;
                self._stamp(context, _log);
                debug(_log.code);
                logger.debug(_log);
                return true;
            },
            audit: function (_audit) {
                _audit.code = _audit.code || _audit.action;
                assert(_audit.code, "missing audit.code");
                _audit.message = _audit.message || _audit.code;
                self._stamp(context, _audit);
                debug(_audit.code);
                bus.emit("api:audit", _audit);
                auditor.info(_audit);
                return true;
            }
        };
        return context;
    }
    _stamp(context, _log) {
        if (!IS_DEV_ENV) {
            _log.serviceId = context.pkg.name;
            _log.serviceVersion = context.pkg.version;
            _log.serviceUUID = context.uuid;
            _log.timestamp = Date.now();
        }
        return _log;
    }
    boot(bootFn) {
        if (this.booted) {
            return;
        }
        bootFn && bootFn.apply(this);
        this.booted = true;
        // publish our first (local) event - since our context is ready
        this.emit_log("api:boot", true, true, { port: this.config.port, environment: NODE_ENV });
    }
    emit_log(code, verbose_context, do_emit, _logged) {
        _logged = _.extend({ code: code,
            name: this.config.name
        }, _logged);
        if (this.context.pkg) {
            _logged.package = this.context.pkg.name;
            _logged.version = this.context.pkg.version;
        }
        if (verbose_context) {
            _logged.plugins = _.keys(this.context.plugins.features);
            _logged.middleware = _.keys(this.context.middleware.features);
        }
        if (do_emit) {
            this.context.bus.emit(_logged.code, _logged);
        }
        this.context.log(_logged);
        return _logged;
    }
    installPlugins(context, config) {
        let features = _.extend({ pipeline: { enabled: true }, openapi: { enabled: true } }, config.features);
        for (let name in features) {
            let options = _.extend({ enabled: true }, config.features[name]);
            let plugin = this.context.plugins.get(name);
            let enabled = (options && options.enabled !== false); // enabled by default
            if (!plugin) {
                context.error({ code: "feature:plugin:missing", plugin: name, message: "plugin not registered" });
                throw new Error("feature:plugin:missing:" + name);
            }
            else if (enabled) {
                delete options.enabled; // clean-up so options can be consumed by plugins more easily
                let _log = { code: "feature:plugin:enabled", plugin: name };
                context.log(_log);
                plugin.install(context, options);
                context.bus.emit(_log.code, { plugin: name, options: options });
            }
            else {
                context.warn({ code: "feature:plugin:disabled", plugin: name, message: "plugin disabled" });
            }
        }
    }
    /**
     * Register Plugin
     * @param {IChassisPlugin} plugin
     */
    registerPlugin(plugin) {
        this.context.trace({ code: "feature:plugin:register", name: plugin.name, message: plugin.title });
        this.plugins.set(plugin);
    }
    /**
     * Register Middleware
     * @param {IChassisPlugin} plugin
     */
    registerFn(plugin) {
        this.context.trace({ code: "feature:fn:register", name: plugin.name, message: plugin.title });
        this.middleware.set(plugin);
    }
    process_guard() {
        let context = this.context;
        // process crashes
        process.on('uncaughtException', function (err) {
            context && context.error({ code: "api:error:uncaughtException", message: err.message, stack: err.stack });
            console.error(err);
            process.exit(1);
        });
        // Promise rejections
        process.on('unhandledRejection', function (err) {
            context && context.error({ code: "api:error:unhandledRejection", message: err.message, stack: err.stack });
            console.error(err);
            process.exit(2);
        });
        context && context.log({ code: "api:process_guard" });
    }
}
exports.Chassis = Chassis;
//# sourceMappingURL=Chassis.js.map