"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
class GQL {
    constructor(_schemas) {
        this.json_schemas = {};
        this.gql_types = {};
        if (_schemas)
            this.schemas(_schemas);
    }
    schemas(_schemas) {
        // generate types
        _.each(_schemas, (s, n) => {
            let name = GQL.$(n);
            this.gql_types[name] = GQL.to_type(name, s);
        });
    }
    types() {
        return this.gql_types;
    }
    /**
     * sanitize `txt` by converting to lower case and replacing non-alpha with `_` (underscore)
     * @param txt
     */
    static $(txt) {
        return txt.toLowerCase().replace(/[\W_]+/g, "_");
    }
    /**
     * construct a named GraphQL query from the schema
     *
     * @param name
     * @param schema
     */
    static query(name, schema) {
        return "query get_" + this.$(name) + " { " + this.to_query(name, schema) + " }";
    }
    /**
     * return ! if `name` is a required field within schema
     *
     * @param name
     * @param schema
     */
    static is_required(name, schema) {
        let required = schema.required || [];
        if (required.indexOf(name))
            return true;
        if (!schema.properties)
            return false;
        for (let fieldname in schema.properties) {
            let prop = schema.properties[fieldname];
            if (prop.required)
                return true;
        }
        return false;
    }
    /**
     * return GraphQL type definition (schema)
     *
     * @param name
     * @param schema
     */
    static to_type(name, schema) {
        name = this.$(name);
        let gql = "type " + name + " {";
        console.log("type: %o", schema.type);
        gql += this.to_type_properties(schema, schema);
        gql += "}";
        return gql;
    }
    static to_type_properties(root, schema) {
        if (schema.type == "array") {
            return this.to_type_properties(schema, schema.items);
        }
        else if (schema.type == "object") {
            let gql = "";
            for (let fieldname in schema.properties) {
                let prop = schema.properties[fieldname];
                let gql_type = this.schema_to_gql_types[prop.type || "string"];
                let non_null = this.is_required(fieldname, schema) ? "!" : "";
                fieldname = this.$(fieldname);
                if (prop.type == "array") {
                    gql += " [" + fieldname + ": [" + gql_type + non_null + "]";
                }
                else {
                    gql += " " + fieldname + ": " + gql_type + non_null;
                }
            }
            return gql;
        }
    }
    /**
     * create argument list for parameters
     *
     * @param args
     */
    static to_args(args) {
        if (!args)
            return "";
        let gql = "(";
        for (let arg in args) {
            gql += " " + arg + ": \"" + args[arg] + "\"";
        }
        gql += ")";
        return gql;
    }
    /**
     * construct a query fragment from the named schema
     *
     * @param name
     * @param schema
     * @param args
     */
    static to_query(name, schema, args) {
        name = this.$(name);
        if (schema.items)
            return this.to_query(name, schema.items);
        let gql = " ";
        gql += name;
        gql += this.to_args(args);
        gql += " {";
        for (let fieldname in schema.properties) {
            let prop = schema.properties[fieldname];
            let type = prop.type;
            switch (type) {
                case "object":
                    gql += this.to_query(fieldname, prop);
                    break;
                case "array":
                    gql += "XX_" + this.to_query(fieldname, prop.items);
                    break;
                default:
                    gql += " " + this.$(fieldname);
            }
        }
        gql += " }";
        return gql;
    }
}
GQL.schema_to_gql_types = {
    "string": "String"
};
exports.default = GQL;
//# sourceMappingURL=GQL.js.map