"use strict";
/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Ventures Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const helpers_1 = require("../helpers/");
const _ = require("lodash");
const JWT_1 = require("../helpers/JWT");
const _1 = require(".");
const DEFAULT_CLAIM_PATHS = ["$.roles", "$.realm_access.roles", "$.resource_access.account.roles"];
/**
 * Authorize JWT
 * -------------
 * Ensure the JWT contains specified claims.
 *
 * Unless disabled, an audit event is emitted.
 *
 */
class jwt_openapi {
    constructor(plugin, openapi) {
        this.plugin = plugin;
        this.openapi = openapi;
        this.name = "api.jwt_openapi";
        this.title = "OpenAPI JWT validation";
    }
    /**
     * generate claims from scopes (jsonata format)
     *
     * @param rules_scopes
     * @returns {{}}
     */
    normalize_scopes(rules_scopes) {
        let all_scopes = [];
        for (let rs in rules_scopes) {
            let scope = rules_scopes[rs];
            if (_.isArray(scope)) {
                all_scopes = all_scopes.concat(scope);
            }
            else if (_.isString(scope)) {
                all_scopes = all_scopes.concat(scope.split(","));
            }
        }
        return all_scopes;
    }
    ;
    // create a set of named jsonata matchers - one for each claim in the path
    claims_matchers(scope_claims, claim_rules) {
        let claims_matchers = {};
        _.each(claim_rules, function (matcher_key) {
            claims_matchers[matcher_key] = scope_claims;
        });
        return claims_matchers;
    }
    security_scheme(rules_type, rules_scheme) {
        // collect the set of rules that match scheme/type ..
        //            let rules = this.find_scheme(options, rules_type, rules_scheme);
        let securityScheme = this.openapi.security.findScheme(rules_type, rules_scheme);
        // we want to fail fast, if configuration doesn't work ...
        assert(securityScheme, "missing scheme matching type=" + rules_type + " & scheme=" + rules_scheme);
        assert(securityScheme.bearerFormat == "JWT", "securityScheme.bearerFormat not JWT");
        return securityScheme;
    }
    /**
     * Returns OpenAPI authorization middleware
     *
     * The middleware is bound to the scopes/claims in the x-feature definition:
     *
     *      scopes:
     *
     *
     * Injected by openapi/paths.ts
     *
     * @param {IChassisContext} context
     * @param oper
     * @returns {(req, res, next) => boolean}
     */
    fn(operation, _options) {
        let options = _.extend({ matchers: DEFAULT_CLAIM_PATHS }, operation.feature, _options);
        let context = operation.context;
        // collect the set of rules that match scheme/type ..
        let securityScheme = this.security_scheme("http", "bearer");
        // find all the scopes for current resource
        let scope_claims = this.normalize_scopes(securityScheme.scopes);
        // which jsonpaths in our JWT hold claims - "$.roles", "$.realm_access.roles"
        // compile jsonpaths into jsonata 'claim' matchers
        let claims_matchers = this.claims_matchers(scope_claims, options.matchers);
        let is_protected = claims_matchers.length > 0;
        let authn = new _1.jwt_authenticate(this.plugin).fn(operation, operation);
        context.log({ code: "api:jwt:protected", message: (is_protected ? "" : "not ") + "protected", protected: is_protected, resource: operation.resource, method: operation.actionId, matchers: _.keys(claims_matchers), expected: scope_claims });
        return function (req, res, next) {
            // un-protected API - so skip checks
            if (!is_protected) {
                next();
                return;
            }
            authn(req, res, () => {
                let jwt = req.jwt;
                if (!jwt) {
                    return helpers_1.Utils.sendError(context, res, { statusCode: 401, code: "api:jwt:missing", message: "missing token" });
                }
                // does our JWT match our required OAUTH claims
                let has_claims = JWT_1.default.hasClaims(jwt, claims_matchers);
                // send an error
                if (!has_claims) {
                    return helpers_1.Utils.sendError(this.context, res, { statusCode: 403, code: "api:jwt::unauthorized", message: "missing claims", resource: operation.resource, method: operation.actionId, expected: scope_claims, jwt: jwt });
                }
                // we're good - continue processing request ...
                this.context.audit({ code: "api:jwt::authorized", message: "authorized", resource: operation.resource, method: operation.actionId, jti: jwt.jti, expected: scope_claims });
                next();
            });
        };
    }
}
exports.default = jwt_openapi;
//# sourceMappingURL=jwt_openapi.js.map