"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const _ = require("lodash");
const APIError_1 = require("../core/APIError");
const Proxy = require("http-proxy-middleware");
const helpers_1 = require("../helpers");
const querystring = require("query-string");
// import { ClientRequest } from "http";
/**
 * HTTP Reverse Proxy
 * ------------------
 * Attach a remote endpoint to an OpenAPI resource
 * The implementation of a HTTP Proxy plugin using the node-http-proxy package.
 *
 * See https://github.com/nodejitsu/node-http-proxy
 *
 * TODO: integrate with Vault to retrieve secrets and inject into egress request via header/params/etc.
 *
 */
class proxy {
    constructor() {
        this.name = "api.proxy";
        this.title = "Reverse proxy";
        this.proxyServer = null;
        /**
         * Create express middleware that forwards a request to a remote endpoint
         *
         * @param context
         * @param options
         * @param proxyPath
         * @returns {Function}
         * @constructor
         */
        this.proxy_middleware = function (context, _options, oper) {
            let options = this._target_options(_options);
            let req_path = oper.resource;
            if (req_path.endsWith("/*")) {
                req_path = req_path.substring(0, req_path.length - 2);
            }
            context.log({ code: "api:proxy:setup", message: oper.resource + " -> " + options.target, req_path: req_path, proxyPath: oper.resource, options: options });
            return (req, res, _next) => {
                let params = _.extend({}, req.query, req.params);
                let vars = _.extend({ resource: oper.resource, params: params });
                try {
                    context.log({ code: "api:proxy:target", message: "", resource: oper.resource, target: options.target });
                    // interpolate vars in target path (template)
                    options.target = helpers_1.Vars.$(options.target, vars);
                }
                catch (e) {
                    context.error({ code: "api:proxy:target:template:failed", message: e.message, resource: oper.resource, vars: vars });
                }
                if (!options.onError)
                    options.onError = (err) => {
                        return helpers_1.Utils.sendError(context, res, { code: "api:proxy:failed", message: err.message || "unknown error", error: err, statusCode: 500 });
                    };
                // Work-around for body-parsers breaking proxy
                if (!options.onProxyReq)
                    options.onProxyReq = (proxyReq, req, _res) => {
                        // inject credentials into proxy request
                        this.reuseJWT(context, options, req, proxyReq);
                        if (!this.useCredentials(context, options, req, proxyReq)) {
                            throw new APIError_1.default(500, { code: "api:proxy:credentials:missing" });
                            return;
                        }
                        // remove mount-path from request
                        if (proxyReq.path.startsWith(req_path)) {
                            proxyReq.path = proxyReq.path.substr(req_path.length);
                            context.log({ code: "api:proxy:target:path", resource: oper.resource, proxyPath: proxyReq.path, req_path: req_path });
                        }
                        // empty body - do nothing
                        if (!req.body || !Object.keys(req.body).length) {
                            return;
                        }
                        // re-write body based on content-type
                        const contentType = proxyReq.getHeader('Content-Type');
                        const writeBody = (bodyData) => {
                            proxyReq.setHeader('Content-Length', Buffer.byteLength(bodyData));
                            proxyReq.write(bodyData);
                        };
                        if (contentType === 'application/json') { // contentType.includes('application/json')
                            writeBody(JSON.stringify(req.body));
                        }
                        if (contentType === 'application/x-www-form-urlencoded') {
                            writeBody(querystring.stringify(req.body));
                        }
                    };
                let proxy = Proxy(options);
                return proxy(req, res, _next);
            };
        };
        /**
         * If enabled, copy the incoming JWT to the proxy request.
         * Otherwise, ensure the JWT is removed.
         *
         * @param context
         * @param options
         * @param req
         * @param proxyReq
         */
        this.reuseJWT = function (context, options, req, proxyReq) {
            if (options.reuseJWT && req.jwt) {
                let token = req.header('authorization');
                proxyReq.setHeader('authorization', token);
                context.log({ code: "api:proxy:jwt:inject", path: proxyReq.path, jti: req.jwt.jti, token: token });
            }
            else {
                context.log({ code: "api:proxy:jwt:redact", path: proxyReq.path, jwt: req.jwt ? true : false });
                proxyReq.removeHeader('authorization');
            }
        };
    }
    /**
     * plugin to proxy requests to a remote HTTP endpoint
     *
     * @param context
     * @param options
     * @returns {TargetEndpoint}
     */
    fn(operation, _options) {
        let context = operation.context;
        assert(operation.resource, "missing operation.resource");
        let options = _.extend({}, operation.feature.proxy, _options);
        // context.log({ code: "api:proxy:config", options: options });
        // wire-up our reverse-proxy - but only once
        // this.setupProxyServer( operation, options );
        // create the endpoint
        return this.proxy_middleware(context, options, operation);
    }
    ;
    _target_options(options) {
        // set defaults - NOTE: different defaults to https://github.com/nodejitsu/node-http-proxy
        let target_options = _.extend({
            target: false, reuseJWT: false, selfHandleResponse: false,
            autoRewrite: true, xfwd: true, prependPath: true,
            ignorePath: true, changeOrigin: true, toProxy: false, logLevel: 'debug'
        }, options);
        return target_options;
    }
    // inject secret from ENV into HEADER
    // see also: https://ec.haxx.se/http-auth.html
    useCredentials(_context, _options, _req, proxyReq) {
        let secretEnvKey = _options.secretEnvKey || "";
        let secretHeader = _options.secretHeader || "authorization";
        let secretHeaderType = _options.secretHeaderType || "raw";
        let secretValue = process.env[secretEnvKey] || false;
        // skip .. if not env-key specific
        if (!secretEnvKey)
            return true;
        let _msg = { code: "api:proxy:credentials", message: secretHeaderType + " " + secretEnvKey,
            headerType: secretHeaderType, secretHeader: secretHeader, secretEnvKey: secretEnvKey,
            secretFound: secretValue ? true : false };
        _context.log(_msg);
        // fail if secret not found
        if (!secretValue) {
            return false;
        }
        // inject secret into header
        switch (secretHeaderType.toLowerCase()) {
            case "raw":
                proxyReq.setHeader(secretHeader, secretValue);
                break;
            case "basic":
                proxyReq.setHeader(secretHeader, "Basic " + secretValue);
                break;
            case "bearer":
                proxyReq.setHeader(secretHeader, "Bearer " + secretValue);
                break;
            default:
                proxyReq.removeHeader(secretHeader);
        }
        return true;
    }
}
exports.default = proxy;
//# sourceMappingURL=proxy.js.map