"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// import { Security } from "./OpenAPI";
const _ = require("lodash");
const assert = require("assert");
// import SecurityChecks from "./SecurityChecks";
// import { Vars } from "../helpers";
/**
 * Represents an OAS3 `operation`. The operation is executed by `middleware` feature.
 *
 * It combines the OAS3 specification with a flexible request/response processing pipeline.
 *
 * The `pipeline` supports `before` pre-processing and `after` post-processing of requests.
 *
 * The `pipeline` provides support for OAS3 `security` policy enforcement.
 *
 * Every stage of the `pipeline` supports logging and auditing.
 *
 * TODO: support request and response validation
 *
 */
const OPER_PARAMS = ["operationId", "summary", "description", "security", "responses", "parameters", "tags"];
class Operation {
    /**
    * Create a new RequestFlow pipeline
    *
    * @param openapi
    * @param operationId
    * @param method
    * @param resource
    * @param _spec
    */
    constructor(context, actionId, resource, _spec) {
        this.context = context;
        this.ix_parameters = { path: {}, body: {}, query: {} };
        this.ix_responses = { 200: {}, 400: {}, 404: {}, 500: {}, default: { description: "default response", content: { "application/json": {} } } };
        this.standardize = function (_oper) {
            let oper = _.extend({ operationId: "unknown", summary: "", description: "",
                security: [], responses: {}, parameters: [], tags: [] }, _.pick(_oper, OPER_PARAMS));
            delete oper['x-feature'];
            delete oper['feature'];
            return oper;
        };
        this.response = function (status) {
            return this.ix_responses[status] || {};
        };
        this.context = context;
        this.spec = this.standardize(_spec);
        this.feature = _spec.feature || {};
        this.featureId = this.feature.featureId || "api.pipeline";
        assert(this.featureId, "missing feature for: " + actionId + " " + resource);
        this.resource = resource.replace("{", ":").replace("}", ""); //transform Swagger {{param}} into Express :param
        this.actionId = actionId;
        let resource_name = resource == "/" ? "root" : resource;
        this.operationId = this.spec.operationId = (_spec.operationId || this.actionId + "-" + resource_name).replace(/[\W_]+/g, "-");
        this.process_responses(this.spec);
    }
    process_responses(spec) {
        let self = this;
        // index paramters by `in` field
        this.ix_parameters = { path: {}, body: {}, query: {} };
        _.each(spec.parameters, (p) => {
            let p_in = self.ix_parameters[p.in] = self.ix_parameters[p.in] || {};
            p_in[p.name] = { required: p.required ? true : false, schema: p.schema || {} };
        });
        // ensure a default response
        if (_.isEmpty(spec.responses)) {
            spec.responses = { default: this.ix_responses.default };
        }
        else {
            // index responses by `status`
            _.each(spec.responses, (r, status) => {
                let r_status = self.ix_responses[status] = self.ix_responses[status] || {};
                _.each(r, (content, mime) => {
                    r_status[mime.toLowerCase()] = { schema: content.schema };
                });
            });
        }
    }
    parameterKeys(type) {
        return _.keys(this.ix_parameters[type]);
    }
    respond(res) {
        let _response = this.response(res.status || "default");
        let content_type = res.headers["content-type"] || false;
        let _payload = _response[content_type] || { code: "api:response:unknown", content_type: content_type };
        res.send(_payload);
        return _payload;
    }
}
exports.Operation = Operation;
//# sourceMappingURL=Operation.js.map