import * as jsonata from "jsonata";
import Vars from "../helpers/Vars"; 
import {IChassisContext, IJWT } from "../index";
import * as _ from "lodash";
import assert = require("assert");
import { Utils } from ".";
import * as jwt from "jsonwebtoken";
import JWTPlugin from "../plugins/jwt";

const BEARER = "bearer ";

export default class JWT {

    public static getHeaders(token: string, headers?: any): any {
        headers = headers || {};
        headers.Authorization = 'Bearer '+token;
        return headers;
    }

    public static getBearerToken (context: IChassisContext, req: any) {
        return this.getBearerTokenOrSendError(context, req);
    }

    public static getJWT(context: IChassisContext, req: any, res?: any): any {

        // extract JWT from headers - or error if missing or invalid
        let token = this.getBearerTokenOrSendError(context,req ,res);
        if (!token) return;
        // decode JWT into JSON
        let decoded = jwt.decode(token);
        return decoded;
    }
    public static getValidJWT(context: IChassisContext, req: any, res?: any): any {
        req.jwt = false;

        let jwt_plugin = context.plugins.get("jwt") as JWTPlugin;
        let has_secretOrPublicKey = (jwt_plugin && jwt_plugin.secretOrCertificate)?true:false

        if (!has_secretOrPublicKey) {
            res && Utils.sendError(context, res, {statusCode: 403, code: "api:oauth:token:invalid", message: "unverified token"});
            return null
        }

        // extract JWT from headers - or error if missing or invalid
        let token = this.getBearerTokenOrSendError(context,req ,res);
        if (!token) {
            return null;
        }
    
        // see https://github.com/auth0/node-jsonwebtoken for options: audience, issuer, jwtid, subject, etc:
        let verified = null;
        let verify_options = _.extend({ algorithms: [jwt_plugin.algo] }, _.pick(context.config.jwt, ["algorithms", "issuer", "jwtid", "maxAge"]) );
        try {
            verified = jwt.verify(token, jwt_plugin.secretOrCertificate, verify_options );
        } catch(e) {
            context.log({ code: "api:oauth:jwt:invalid", error: e, options: verify_options, secretOrCertificate: jwt_plugin.secretOrCertificate });
        }
        if (!verified) {
            res && Utils.sendError(context, res, {statusCode: 403, code: "api:oauth:token:corrupt", message: "invalid token"});
            return null;
        }

        // cache the decoded token in the request
        let decoded = jwt.decode(token);
        return req.jwt = decoded || {};
    };

    /**
     * Check for internal consistency within the JWT
     *  not expiry date/time
     *  not before valid date/time
     * 
     * @param context 
     * @param jwt 
     * @param res 
     */
    sanityCheckOrSendError(context: IChassisContext, jwt: IJWT, res?: any):boolean {
        let now = Date.now() / 1000 // now in seconds;

        // check JWT: has not expired
        if (now > jwt.exp) {
            return res?false:Utils.sendError(context, res, { statusCode: 403, code: "api:jwt:expired", message: "expired", expires: new Date(jwt.exp*1000), jwt: jwt });
        }

        // check JWT: not yet issued
        if ( now < jwt.nbf ) {
            return res?false:Utils.sendError(context, res, { statusCode: 403, code: "api:jwt:nbf", message: "not active", nbf: new Date(jwt.nbf*1000), jwt: jwt });
        }
        return true
    }

    public static getBearerTokenOrSendError (context: IChassisContext, req: any, res?: any) {
        // check Authorization Bearer header exists
        let header = req.headers.authorization || "";
        let has_token = header.toLowerCase().indexOf(BEARER)==0;
        let token = has_token?header.substring(BEARER.length).trim():"";
        if (token) return token;

        // auto-response is optional ...
        res && Utils.sendError(context, res, {statusCode: 401, code: "api:oauth:token:missing", message: "not-authenticated"});
        return null;
    };


    /**
     * JWT must contain a non-empty `subject`. 
     * If claims are empty or not specified, then we return true
     * Otherwise, 
     * @param jwt 
     * @param claims 
     */
    public static hasClaims(jwt: any, claims?: any, context?: IChassisContext) {
        if (!jwt || !jwt.sub) {
            context && context.log( { code: "api:jwt:subject", message: "missing jwt.sub", jwt: jwt });
            return false;
        }
        if (_.isEmpty(claims)) {
            context && context.log( { code: "api:jwt:sub", message: "jwt sub"});
            return true; // no claims - so 'allow'
        }
    
        let is_allowed = true;

        _.each(claims, function(claim: any, pattern: string) {
            assert( _.isString(claim) || _.isArray(claim), "invalid claim type");

            if (pattern.indexOf("$.") != 0) pattern = "$."+pattern; // uplift a simple path/pattern into simple jsonata matcher
            console.log("pattern: %s -> %s", typeof pattern, pattern);
    
            let matcher = jsonata(pattern); // create evaluator for claim
            let matched: any = matcher.evaluate(jwt); // evaluate claim against JWT
            is_allowed = is_allowed && matched; // we must have some matches

            // check if the claims were found
            if (_.isArray(claim)) {
                let matches = Vars.intersect( matched, claim);
                console.log("matches: %o", matches, is_allowed);
                is_allowed = is_allowed && matches.length>0 ;
            } else {
                claim = claim.toString();
                matched = matched.toString();
                is_allowed = is_allowed && matched == claim;
                console.log("matched: %o = %o => %s", matched, claim, is_allowed);
            }
        });
        context && context.log( { code: "api:jwt:claims", message: "jwt claims", claims: claims, success: is_allowed });
        return is_allowed;
    }

}