/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Ventures Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */

// import * as assert from "assert";
import * as _ from "lodash";

import * as jwt from "jsonwebtoken";
import {IChassisContext, IOperation, IChassisMiddleware, Utils, JWT } from "../index";
import { Request, Response } from "express";
import JWTPlugin from "../plugins/jwt";

/*

curl -v --data "grant_type=password&client_id=test-lab&client_secret=abcdefgg-0a50-4929-967a-178bf454314a&username=test-lab&password=$OAUTH_PASS" https://oauth.pro/auth/realms/example/protocol/openid-connect/token

curl -v https://oauth.pro/auth/realms/example/protocol/openid-connect/certs

 */

/**
 * Authenticate JWT
 * ----------------
 * Ensures a JWT is correctly formed and signed by the trusted KeycloakIdP.
 *
 * Loads the server public certificate from file or URL.
 *
 * Can enforce mandatory claims for all resources.
 *
 * @type {{name: string, title: string, description: string, usecase: [string], _loadPublicKey: module.exports._loadPublicKey, fn: module.exports.fn}}
 */

export default class jwt_authenticate implements IChassisMiddleware {

    name = "api.jwt_authenticate";
    title = "ensure a JWT is authentic (correctly formed and signed)";

    constructor(protected plugin: JWTPlugin) {
    }

    fn(operation: IOperation, _options: any): Function {

        let context = operation.context;

        return function(req: Request, res: Response, next: Function ) {
            if ( (req as any).jwt ) {
                return next(); // no need to run again
            }

            let authenticated = this.authenticate(context, operation, req, res, this.plugin.secretOrCertificate);
            if (!authenticated) {
                return Utils.sendError(context, res, {statusCode: 401, code: "api:jwt:not-authenticated", message: "missing or invalid token"});
            }

            next();
        };
    }

    public static authenticate(context: IChassisContext, op: IOperation, req: Request, res: Response, secretOrCertificate?: string): boolean {
        let req_uuid = res.getHeader("x-request-id");

        let token = JWT.getBearerTokenOrSendError(context,req);
        if (!token) {
            context.log({ code:  "api:jwt:missing", message: "missing token", resource: op.resource, request_id: req_uuid });
            return false;
        }


        // validate and decode
        if (secretOrCertificate) {
            (req as any).jwt = jwt.verify(token, secretOrCertificate);
        } else {
            (req as any).jwt = jwt.decode(token);
            let log = { code:  "api:jwt:decoded", message: "NOT VERIFIED", resource: op.resource, request_id: req_uuid };
            context.warn(log);
            context.audit(log);
        }
        return (req as any).jwt?true:false;

    }
}
