import * as assert from "assert";
import * as _ from "lodash";
import {IChassisFeature, IOperation, IChassisContext} from "../interfaces";

import APIError from "../core/APIError";
import { Handler } from "express-serve-static-core";
import * as Proxy from "http-proxy-middleware";
import { Request, Response, NextFunction } from "express";
import { Utils, Vars } from "../helpers";
import * as querystring from "query-string";
// import { ClientRequest } from "http";

/**
 * HTTP Reverse Proxy
 * ------------------
 * Attach a remote endpoint to an OpenAPI resource
 * The implementation of a HTTP Proxy plugin using the node-http-proxy package.
 *
 * See https://github.com/nodejitsu/node-http-proxy
 *
 * TODO: integrate with Vault to retrieve secrets and inject into egress request via header/params/etc.
 *
 */

export default class proxy implements IChassisFeature {

    name = "api.proxy";
    title = "Reverse proxy";

    proxyServer: any = null;

    constructor() {

    }

    /**
     * plugin to proxy requests to a remote HTTP endpoint
     *
     * @param context
     * @param options
     * @returns {TargetEndpoint}
     */

    fn(operation: IOperation, _options: any): Function {
        let context = operation.context;

        assert(operation.resource, "missing operation.resource");

        let options = _.extend({}, operation.feature.proxy, _options);

        // context.log({ code: "api:proxy:config", options: options });

        // wire-up our reverse-proxy - but only once
        // this.setupProxyServer( operation, options );

        // create the endpoint
        return this.proxy_middleware(context, options, operation);
    };

    /**
     * Create express middleware that forwards a request to a remote endpoint
     *
     * @param context
     * @param options
     * @param proxyPath
     * @returns {Function}
     * @constructor
     */

    proxy_middleware = function (context: IChassisContext, _options: any , oper: IOperation): Handler {
        let options = this._target_options(_options);

        let req_path = oper.resource;
        if (req_path.endsWith("/*")) {
            req_path = req_path.substring(0,req_path.length-2);
        }

        context.log({ code: "api:proxy:setup", message: oper.resource+" -> "+options.target, req_path: req_path, proxyPath: oper.resource, options: options });

        return (req: Request, res: Response, _next: NextFunction) => {
            let params = _.extend({}, req.query, req.params)
            let vars = _.extend({ resource: oper.resource, params: params });

            try {
                context.log( { code: "api:proxy:target", message: "", resource: oper.resource, target: options.target });
                // interpolate vars in target path (template)
                options.target = Vars.$(options.target, vars);
            } catch(e) {
                context.error( { code: "api:proxy:target:template:failed", message: e.message, resource: oper.resource, vars: vars });
            }

            if (!options.onError) options.onError = (err) => {
                return Utils.sendError(context, res, { code: "api:proxy:failed", message: err.message || "unknown error", error: err, statusCode: 500 });
            }

            // Work-around for body-parsers breaking proxy
            if (!options.onProxyReq) options.onProxyReq = (proxyReq: any, req: Request, _res) => {

                // inject credentials into proxy request
                this.reuseJWT(context, options, req, proxyReq);
                if ( !this.useCredentials(context, options, req, proxyReq) ) {
                    throw new APIError(500, { code: "api:proxy:credentials:missing" });
                    return;
                }

                // remove mount-path from request
                if (proxyReq.path.startsWith(req_path)) {
                    (proxyReq as any).path = proxyReq.path.substr(req_path.length);
                    context.log( { code: "api:proxy:target:path", resource: oper.resource, proxyPath: (proxyReq as any).path, req_path: req_path });
                }

                // empty body - do nothing
                if (!req.body || !Object.keys(req.body).length) {
                    return;
                }

                // re-write body based on content-type
                const contentType = proxyReq.getHeader('Content-Type');
                const writeBody = (bodyData: string) => {
                    proxyReq.setHeader('Content-Length', Buffer.byteLength(bodyData));
                    proxyReq.write(bodyData);
                };
                
                if (contentType === 'application/json') {  // contentType.includes('application/json')
                    writeBody(JSON.stringify(req.body));
                }
                
                if (contentType === 'application/x-www-form-urlencoded') {
                    writeBody(querystring.stringify(req.body));
                }
            }

            let proxy = Proxy(options);

            return proxy(req, res, _next);
        }
    }

    _target_options(options: any) {
        
        // set defaults - NOTE: different defaults to https://github.com/nodejitsu/node-http-proxy
        let target_options = _.extend({ 
            target: false, reuseJWT: false, selfHandleResponse: false, 
            autoRewrite: true, xfwd: true, prependPath: true, 
            ignorePath: true , changeOrigin: true, toProxy: false, logLevel: 'debug'
        }, options);
        
        return target_options;
    }

    /**
     * If enabled, copy the incoming JWT to the proxy request.
     * Otherwise, ensure the JWT is removed.
     *
     * @param context
     * @param options
     * @param req
     * @param proxyReq
     */
    reuseJWT = function(context: IChassisContext, options: any , req: any, proxyReq: any ) {

        if (options.reuseJWT && req.jwt) {
            let token = req.header('authorization');
            proxyReq.setHeader('authorization', token);
            context.log({ code: "api:proxy:jwt:inject", path: proxyReq.path, jti: req.jwt.jti, token: token });
        } else {
            context.log({ code: "api:proxy:jwt:redact", path: proxyReq.path, jwt: req.jwt?true:false });
            proxyReq.removeHeader('authorization');
        }
    };

    // inject secret from ENV into HEADER
    // see also: https://ec.haxx.se/http-auth.html
    useCredentials(_context: IChassisContext, _options: any , _req: any, proxyReq: any): boolean {
        let secretEnvKey = _options.secretEnvKey || "";
        let secretHeader = _options.secretHeader || "authorization";
        let secretHeaderType = _options.secretHeaderType || "raw";
        let secretValue = process.env[secretEnvKey] || false;

        // skip .. if not env-key specific
        if (!secretEnvKey) return true;

        let _msg = { code: "api:proxy:credentials", message: secretHeaderType+" "+secretEnvKey, 
            headerType: secretHeaderType, secretHeader: secretHeader, secretEnvKey: secretEnvKey, 
            secretFound: secretValue?true:false };
        _context.log(_msg);

        // fail if secret not found
        if ( !secretValue ) {
            return false;
        }

        // inject secret into header
        switch(secretHeaderType.toLowerCase()) {
            case "raw":
                proxyReq.setHeader(secretHeader, secretValue);
            break;
            case "basic":
                proxyReq.setHeader(secretHeader, "Basic "+secretValue);
            break;
            case "bearer":
                proxyReq.setHeader(secretHeader, "Bearer "+secretValue);
            break;
            default:
                proxyReq.removeHeader(secretHeader);
        }
        return true;
    }
}
