/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Ventures Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */
import * as _ from "lodash";
import { IOpenAPIv3 } from "../interfaces";
import { Vars } from "../helpers";
import { Operation } from "./Operation";

const REF_ID_PREFIX = "#/components/schemas/";
const X_SCHEMA_SRC = "x-schema-source";

export class Schemas {
    // context: IChassisContext;
    schemas: any = {};

    constructor(spec?: IOpenAPIv3) {
        spec && this.init(spec);
    }

    /** parse OpenAPI specifications */

    public init(spec: IOpenAPIv3): Schemas {
        if (spec.components && spec.components.schemas) {
            _.each(spec.components.schemas, (schema, schema_id) => {
                this.schema(schema, schema_id);
            });
        }

        _.each(spec.paths, (resource, path) => {
            this.resource(resource, path);
        });
        return this;
    }

    public operation(op: Operation) {
        this.method( op.spec, op.actionId, op.resource);
    }

    resource(resource: any, path: string) {
        this.methods(resource, path);
    }

    methods(methods: any, path: string) {
        _.each(methods, (oper, method) => {
            this.method(oper, method, path);
        });
    }

    method(oper: any, method: string, path: string) {
        let method_ref_id = (method+Vars.sanitize(path, "-")).toLocaleLowerCase();
        // console.log("method_ref_id: (%s) %o", path, method_ref_id);

        this.parameters_schemas(oper.parameters, method_ref_id+"-parameters");
        if (oper.requestBody && oper.requestBody.content) {
            this.requestBody_schema( oper.requestBody.content, method_ref_id+"-requestBody" );
        }
        this.responses_schemas( oper.responses, method_ref_id+"-response" );
    }

    // 
    requestBody_schema(requestBody: any, op_ref_id: string) {
        _.each(requestBody.content, (content, content_type) => {
            let ref_id = op_ref_id+"-"+content_type;
            if (content.schema) {
                let schema = _.extend( { title: content.summary, description: content.description }, content.schema);
                schema[X_SCHEMA_SRC] = "request";
                content.schema = this.schema(schema, ref_id);
            }
        });
    }

    parameters_schemas(parameters: any, op_ref_id: string) {
        _.each(parameters, (param, _name) => {
            let ref_id = op_ref_id+"-"+param.name
            // console.log("path_ref_id: %o", ref_id);
            if (param.schema) {
                let schema = _.extend( { title: Vars.humanize(param.name), description: param.description }, param.schema);
                schema[X_SCHEMA_SRC] = "parameters";
                param.schema = this.schema( schema, ref_id );
            }
        });
    }

    responses_schemas(responses: any, op_ref_id: string) {
        _.each(responses, (resp, status_code) => {
            let ref_id = op_ref_id+"-"+status_code;
            // console.log("responses_ref_id: [%s] %o", status_code, ref_id );
            this.response_schema( status_code, resp, ref_id );
        });
    }   

    response_schema(_status_code: string, response: any, method_ref_id: string) {
        // update if schema set, read otherwise
        _.each(response.content, (response_type: any, _content_type: string) =>{
            if (response_type && response_type.schema) {
                let schema = _.extend( { title: response_type.summary, description: response_type.description }, response_type.schema);
                schema[X_SCHEMA_SRC] = "response";
                // console .log("response-content: %s [ %s ] -> %j", method_ref_id, _content_type, response_type);
                response_type.schema = this.schema(schema, method_ref_id);
            }
        })
        return response;
    }

    _id(id: string) {
        if (id.startsWith(REF_ID_PREFIX)) {
            id = id.substring(REF_ID_PREFIX.length);
        }
        return id;
    }

    schema(_schema: any, ref_id?: string): any {
        if (!_schema) return;
        if (_schema.$ref) {
            console.error("ref-schema skipped: "+_schema.$ref);
            return { $ref: _schema.$ref };
        }

        let schema = _.extend({}, _schema);

        if (!schema[X_SCHEMA_SRC]) {
            schema[X_SCHEMA_SRC] = "default";
        }

        let org_id, id = schema.$id || schema.$ref || ref_id;

        // local name
        if (id.startsWith(REF_ID_PREFIX)) {
            id = id.substring(REF_ID_PREFIX.length)
        }
        if (!id) throw new Error("Missing schema id: "+org_id);

        schema.$id = id;
        if (this.schemas[id]) {
            console.error("duplicate schema: "+id);
        } else {
            // console.log("add-schema: %s ==> %o", id, Vars.clean(schema) );
            this.schemas[id] = schema;
        }

        _.each(schema.properties, (v: any,k)=>{
            if (_.isObject(v)) {
                (v as any).title = (v as any).title || Vars.humanize(k);
            }
        });

        // schema.properties =_.orderBy(schema.properties, ["title"], ['desc']);

        return { $ref: REF_ID_PREFIX+id };
    }
    
    get(id: string) {
        id = this._id(id);
        let schema = _.extend({}, this.schemas[id]);
        return schema;
    }

    strip_x(schema: any): any {
        _.each(schema, (_v,k) => {
            if (k.startsWith("x-")) delete schema[k];
        });
        return schema;
    }

    deref(org_schema: any, _options?: any): any {
        let options = _.extend({ id: false }, _options);

        if ( _.isString(org_schema) ) {
            let id = org_schema;
            org_schema = this.get(id);
            if (!options.id) delete org_schema["$id"];
            if (options.ref) org_schema["$ref"] = id;
        }
        console.log("deref: %j", org_schema );

        let schema = {};
        _.each(org_schema, (v,k) => {
            if (k == "$ref" && v.startsWith(REF_ID_PREFIX)) {
                let id = this._id(v);
                let ref = this.deref(id, options);
                _.extend(schema, ref);
                if (options.id) schema["$id"] = id;
                if (options.ref) schema["$ref"] = v;
               console.log("REF$: %j (%j) -> %j == %j", id, v, schema, options);
            } else if ( _.isObject(v)) {
                let ref = this.deref(v, options);
                this.strip_x(ref);
                schema[k] = ref;
                console.log("deref: %j -> %j", k, ref );
            } else {
                schema[k] = v;
            }
        });
        return schema;
    }

    all(_options?: any): any {
        let options = _.extend({ id: false, ref: false, source: "" }, _options);
        let schemas = {};
        let source_types = options.source.split(",");
        _.each(this.schemas, (v,k) => {
            if (!options.source || source_types.indexOf(v[X_SCHEMA_SRC])>=0 ) {
                schemas[k] = this.deref(v, options);
            }
        });
        return schemas;
    }

    refs():any {

        let org_schema = this.all( { id: true, ref: true } );

        let refs = [];

        function walk(schema: any) {
            _.each(schema, (v) => {
                refs.push( { title: v.title, $id: v.$id, $ref: v.$ref});
                if (v.items) {
                    walk(v.items);
                }
            });
        }

        walk(org_schema);
        return refs;
    }
}